/*
 * This file is part of the Dash-To-Panel extension for Gnome 3
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Credits:
 * This file is based on code from the Dash to Dock extension by micheleg
 * and code from the Taskbar extension by Zorin OS
 * Some code was also adapted from the upstream Gnome Shell source code.
 */

#dashtopanelTaskbar .dash-item-container > StWidget,
.dashtopanelMainPanel .dash-item-container .show-apps {
	margin: 0;
	padding: 0;
}

#dashtopanelScrollview .overview-tile,
.dashtopanelMainPanel .overview-tile {
	background: none;
}

#dashtopanelScrollview .overview-tile .overview-label {
	/* must match TITLE_RIGHT_PADDING in apppicons.js */
	padding-right: 8px;
	text-align: left;
}

#dashtopanelScrollview .overview-tile:hover .dtp-container,
#dashtopanelScrollview .overview-tile:focus .dtp-container {
	background-color: rgba(238, 238, 236, 0.1);
	border-radius: 0px;
}

#dashtopanelScrollview .overview-tile:hover .dtp-container > .dtp-dots-container,
#dashtopanelScrollview .overview-tile:focus .dtp-container > .dtp-dots-container {
	border-radius: 0px;
}

.dashtopanelMainPanel .dash-item-container .show-apps:hover {
	background-color: rgba(238, 238, 238, 0.1);
}
.dashtopanelMainPanel .dash-item-container .show-apps .overview-icon {
	color: #FFF;
}
.dashtopanelMainPanel .dash-item-container .show-apps:hover .overview-icon {
	background: none;
}

#dashtopanelTaskbar .dash-item-container .overview-tile:hover,
#dashtopanelTaskbar .dash-item-container .overview-tile .dtp-container .overview-icon,
#dashtopanelScrollview .overview-tile:hover .dtp-container.no-highlight,
#dashtopanelScrollview .overview-tile:focus .dtp-container.no-highlight {
	background: none;
}

#dashtopanelScrollview .overview-tile:active .dtp-container {
	background-color: rgba(238, 238, 236, 0.18);
}

#dashtopanelScrollview .overview-tile .favorite {
	background-color: rgba(80, 150, 255, 0.4);
}

#dashtopanelTaskbar .scrollview-fade {
	background-gradient-end: rgba(0, 0, 0, 0);
}

.dashtopanelSecondaryMenu {
	max-width: 400px;
}

.dashtopanelMainPanel.vertical .panel-button {
	text-align: center;
}

.dashtopanelMainPanel.vertical .panel-button.vertical *,
.dashtopanelMainPanel.vertical .panel-button.clock-display * {
	padding: 0;
	margin: 0;
}

.dashtopanelMainPanel.vertical .panel-button > *,
.dashtopanelMainPanel.vertical .panel-button.vertical > *,
.dashtopanelMainPanel.vertical .panel-button.vertical .system-status-icon,
.dashtopanelMainPanel.vertical .panel-button.clock-display > *,
.dashtopanelMainPanel.vertical .panel-button.clock-display .clock {
	padding: 8px 0;
}

.dashtopanelMainPanel.vertical .panel-button.clock-display {
	-natural-hpadding: 0;
	-minimum-hpadding: 0;
}

#dashtopanelThumbnailList {
	spacing: 0em;
	padding: 0 1em;
}

#dashtopanelThumbnailList .popup-menu-item {
	padding: 0;
	border-radius: 5px;
	spacing: 0;
}

#dashtopanelThumbnailList .window-box {
	padding: 0;
	spacing: 0;
}

#dashtopanelThumbnailList .preview-window-title {
	padding-top: 1em;
}

.popup-menu.panel-menu {
	margin-bottom: 0;
}

#panel #panelLeft, #panel #panelCenter {
	spacing: 0px;
}

.showdesktop-button-dark-hovered {
	background-color: rgba(200, 200, 200, .4);
}

.showdesktop-button-light-hovered {
	background-color: rgba(55, 55, 55, .4);
}

#dashtopanelScrollview .badge {
	color: rgba(255, 255, 255, 1);
	padding: 0.2em 0.5em;
	border-radius: 1em;
	font-weight: bold;
	text-align: center;
	margin: 0 0 0 2px;
}

#dashtopanelScrollview .number-overlay {
	background-color: rgba(0,0,0,0.8);
}

#dashtopanelScrollview .notification-badge {
	background-color: rgba(214, 0, 0, 0.8);
}

#dashtopanelScrollview .progress-bar {
	/* Customization of the progress bar style, e.g.:
	-progress-bar-background: rgba(0.8, 0.8, 0.8, 1);
	-progress-bar-border: rgba(0.9, 0.9, 0.9, 1); */
}

.symbolic-icon-style {
	-st-icon-style: symbolic;
}


/* border radius, grrr no css variables in ST */
#uiGroup.br4 .dashtopanelPanel.dock,
#uiGroup.br4 .dashtopanelPanel.dock .dashtopanelMainPanel,
#uiGroup.br4 .show-apps,
#uiGroup.br4 .dtp-container,
#uiGroup.br4 .dtp-dots-container,
#uiGroup.br4 #preview-menu, 
#uiGroup.br4 #preview-menu .preview-header-box,
#uiGroup.br4 #preview-menu .preview-container,
#uiGroup.br4 #preview-menu .preview-close-btn-container { 
	border-radius: 4px !important;
	overflow: hidden !important;
}
#uiGroup.br4 .dashtopanelPanel.dock .showdesktop-button { 
	border-radius: 0 4px 4px 0 !important;
}

#uiGroup.br8 .dashtopanelPanel.dock,
#uiGroup.br8 .dashtopanelPanel.dock .dashtopanelMainPanel,
#uiGroup.br8 .show-apps,
#uiGroup.br8 .dtp-container,
#uiGroup.br8 .dtp-dots-container,
#uiGroup.br8 #preview-menu, 
#uiGroup.br8 #preview-menu .preview-header-box,
#uiGroup.br8 #preview-menu .preview-container,
#uiGroup.br8 #preview-menu .preview-close-btn-container { 
	border-radius: 8px !important;
	overflow: hidden !important;
}
#uiGroup.br8 .dashtopanelPanel.dock .showdesktop-button { 
	border-radius: 0 8px 8px 0 !important;
}

#uiGroup.br12 .dashtopanelPanel.dock,
#uiGroup.br12 .dashtopanelPanel.dock .dashtopanelMainPanel,
#uiGroup.br12 .show-apps,
#uiGroup.br12 .dtp-container,
#uiGroup.br12 .dtp-dots-container,
#uiGroup.br12 #preview-menu, 
#uiGroup.br12 #preview-menu .preview-header-box,
#uiGroup.br12 #preview-menu .preview-container,
#uiGroup.br12 #preview-menu .preview-close-btn-container { 
	border-radius: 12px !important;
	overflow: hidden !important;
}
#uiGroup.br12 .dashtopanelPanel.dock .showdesktop-button { 
	border-radius: 0 12px 12px 0 !important;
}

#uiGroup.br16 .dashtopanelPanel.dock,
#uiGroup.br16 .dashtopanelPanel.dock .dashtopanelMainPanel,
#uiGroup.br16 .show-apps,
#uiGroup.br16 .dtp-container,
#uiGroup.br16 .dtp-dots-container,
#uiGroup.br16 #preview-menu, 
#uiGroup.br16 #preview-menu .preview-header-box,
#uiGroup.br16 #preview-menu .preview-container,
#uiGroup.br16 #preview-menu .preview-close-btn-container { 
	border-radius: 16px !important;
	overflow: hidden !important;
}
#uiGroup.br16 .dashtopanelPanel.dock .showdesktop-button { 
	border-radius: 0 16px 16px 0 !important;
}

#uiGroup.br20 .dashtopanelPanel.dock,
#uiGroup.br20 .dashtopanelPanel.dock .dashtopanelMainPanel,
#uiGroup.br20 .show-apps,
#uiGroup.br20 .dtp-container,
#uiGroup.br20 .dtp-dots-container,
#uiGroup.br20 #preview-menu, 
#uiGroup.br20 #preview-menu .preview-header-box,
#uiGroup.br20 #preview-menu .preview-container,
#uiGroup.br20 #preview-menu .preview-close-btn-container { 
	border-radius: 20px !important;
	overflow: hidden !important;
}
#uiGroup.br20 .dashtopanelPanel.dock .showdesktop-button { 
	border-radius: 0 20px 20px 0 !important;
}
